from __future__ import annotations

from typing import TYPE_CHECKING

from aiogram import Bot, F, Router, html
from aiogram.filters.callback_data import CallbackData
from aiogram.fsm.state import State, StatesGroup
from aiogram.types import Message
from aiogram.utils.keyboard import InlineKeyboardBuilder

from raito.plugins.commands import description
from raito.plugins.commands.flags import hidden
from raito.plugins.commands.registration import register_bot_commands
from raito.plugins.keyboards import dynamic
from raito.plugins.roles.data import RoleData
from raito.plugins.roles.roles import ADMINISTRATOR, DEVELOPER, OWNER
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram.fsm.context import FSMContext
    from aiogram.types import CallbackQuery

    from raito.core.raito import Raito

router = Router(name="raito.roles.assign")


class AssignRoleCallback(CallbackData, prefix="rt_assign_role"):  # type: ignore[call-arg]
    """Callback data for assigning roles."""

    role_slug: str


class AssignRoleGroup(StatesGroup):
    """State group for assigning roles."""

    user_id = State()


@dynamic(2)
def roles_list_markup(builder: InlineKeyboardBuilder, roles: list[RoleData]) -> None:
    for role in roles:
        builder.button(
            text=role.emoji + " " + role.name,
            callback_data=AssignRoleCallback(role_slug=role.slug),
        )


@router.message(RaitoCommand("roles", "assign"), DEVELOPER | OWNER | ADMINISTRATOR)
@description("Assigns a role to a user")
@hidden
async def show_roles(message: Message, raito: Raito) -> None:
    await message.answer(
        "🎭 Select role to assign:",
        reply_markup=roles_list_markup(raito.role_manager.available_roles),
    )


@router.callback_query(AssignRoleCallback.filter(), DEVELOPER | OWNER | ADMINISTRATOR)
async def store_role(
    query: CallbackQuery,
    state: FSMContext,
    callback_data: AssignRoleCallback,
    raito: Raito,
) -> None:
    if not query.bot:
        await query.answer("🚫 Bot not found", show_alert=True)
        return
    if not isinstance(query.message, Message):
        await query.answer("🚫 Invalid message", show_alert=True)
        return

    role = raito.role_manager.get_role_data(callback_data.role_slug)
    await state.update_data(rt_selected_role=role.slug)
    await state.set_state(AssignRoleGroup.user_id)

    chat_id = query.message.chat.id
    await query.bot.send_message(
        chat_id=chat_id,
        text=f"{html.bold(role.label)}\n\n{html.blockquote(role.description)}",
        parse_mode="HTML",
    )
    await query.bot.send_message(chat_id=chat_id, text="👤 Enter user ID:")


@router.message(
    AssignRoleGroup.user_id,
    F.text and F.text.isdigit(),
    DEVELOPER | OWNER | ADMINISTRATOR,
)
async def assign_role(message: Message, raito: Raito, state: FSMContext, bot: Bot) -> None:
    data = await state.get_data()
    role_slug = data.get("rt_selected_role")
    if role_slug is None:
        await message.answer("🚫 Role not selected")
        return
    if not message.from_user:
        await message.answer("🚫 User not found")
        return
    if not message.text or not message.text.isdigit():
        await message.answer("🚫 Invalid user ID")
        return
    if not message.bot:
        await message.answer("🚫 Bot instance not found")
        return

    await state.update_data(rt_selected_role=None)
    await state.set_state()

    role = raito.role_manager.get_role_data(role_slug)
    try:
        await raito.role_manager.assign_role(
            message.bot.id,
            message.from_user.id,
            int(message.text),
            role.slug,
        )
    except PermissionError:
        await message.answer("🚫 Permission denied")
        return

    await message.answer(f"❇️ User assigned to {html.bold(role.label)}", parse_mode="HTML")

    handlers = []
    for loader in raito.router_manager.loaders.values():
        handlers.extend(loader.router.message.handlers)

    await register_bot_commands(raito.role_manager, bot, handlers, raito.locales)



import asyncio
from collections.abc import Sequence

from aiogram.dispatcher.event.handler import CallbackType
from aiogram.fsm.storage.base import StorageKey
from aiogram.types import Message
from typing_extensions import NamedTuple

__all__ = ("ConversationRegistry",)


class ConversationData(NamedTuple):
    """Container for an active conversation.

    Stores the Future object awaiting a message and the filters to apply.

    :param future: asyncio.Future that will hold the incoming Message
    :param filters: Sequence of CallbackType filters to validate the message
    """

    future: asyncio.Future[Message]
    filters: Sequence[CallbackType]


class ConversationRegistry:
    """Registry for managing active conversations with users.

    This class allows setting up a "wait for message" scenario where
    a handler can pause and wait for a specific message from a user,
    optionally filtered by aiogram filters.
    """

    STATE = "raito__conversation"

    def __init__(self) -> None:
        """Initialize the conversation registry."""
        self._conversations: dict[StorageKey, ConversationData] = {}

    def listen(self, key: StorageKey, *filters: CallbackType) -> asyncio.Future[Message]:
        """Start listening for a message with a specific StorageKey.

        :param key: StorageKey identifying the conversation (user/chat/bot)
        :param filters: Optional filters to apply when the message arrives
        :return: Future that will resolve with the Message when received
        """
        future = asyncio.get_running_loop().create_future()
        self._conversations[key] = ConversationData(future, filters)
        return future

    def get_filters(self, key: StorageKey) -> Sequence[CallbackType] | None:
        """Get the filters associated with an active conversation.

        :param key: StorageKey identifying the conversation
        :return: Sequence of CallbackType filters or None if no conversation exists
        """
        data = self._conversations.get(key)
        return data.filters if data else None

    def resolve(self, key: StorageKey, message: Message) -> None:
        """Complete the conversation with a received message.

        :param key: StorageKey identifying the conversation
        :param message: Message object that satisfies the filters
        """
        data = self._conversations.pop(key, None)
        if data and not data.future.done():
            data.future.set_result(message)

    def cancel(self, key: StorageKey) -> None:
        """Cancel an active conversation.

        Cancels the Future and removes the conversation from the registry.

        :param key: StorageKey identifying the conversation
        """
        data = self._conversations.pop(key, None)
        if data and not data.future.done():
            data.future.cancel()



from __future__ import annotations

import os
from html import escape
from typing import TYPE_CHECKING

from aiogram import F, Router, html
from aiogram.filters import CommandObject
from aiogram.fsm.state import State, StatesGroup

from raito.plugins.commands import description, hidden
from raito.plugins.roles.roles import DEVELOPER
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram.fsm.context import FSMContext
    from aiogram.types import Message

router = Router(name="raito.system.bash")


class BashGroup(StatesGroup):
    expression = State()


async def _execute_expression(message: Message, text: str) -> None:
    result = os.popen(text).read()
    await message.answer(text=html.pre(escape(result)), parse_mode="HTML")


@router.message(RaitoCommand("bash", "sh"), DEVELOPER)
@description("Execute expression in commandline")
@hidden
async def bash_handler(message: Message, state: FSMContext, command: CommandObject) -> None:
    if not command.args:
        await message.answer(text="📦 Enter expression:")
        await state.set_state(BashGroup.expression)
        return

    await _execute_expression(message, command.args)


@router.message(BashGroup.expression, F.text, DEVELOPER)
async def execute_expression(message: Message, state: FSMContext) -> None:
    await state.clear()

    if not message.text:
        await message.answer(text="⚠️ Expression cannot be empty")
        return

    await _execute_expression(message, message.text)



from aiogram.dispatcher.event.handler import CallbackType, FilterObject, HandlerObject
from aiogram.filters import Filter
from aiogram.types import TelegramObject


async def call_filters(
    event: TelegramObject,
    data: dict,
    *filters: CallbackType,
) -> bool:
    """Run a sequence of filters for a Telegram event.

    :param event: Telegram update object.
    :param data: Aiogram context dictionary.
    :param filters: Aiogram filters to apply.
    :return: True if all filters pass, False otherwise.
    :raises RuntimeError: If `handler` is missing from data.
    """
    handler_object: HandlerObject | None = data.get("handler")
    if handler_object is None:
        raise RuntimeError("Handler object not found")

    for f in filters:
        if isinstance(f, Filter):
            f.update_handler_flags(flags=handler_object.flags)

    for f in filters:
        obj = FilterObject(callback=f)
        if not await obj.call(event, **data):
            return False

    return True



from __future__ import annotations

import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

import psutil
from aiogram import Router, html

from raito.plugins.commands import description, hidden
from raito.plugins.roles.roles import DEVELOPER
from raito.utils.filters import RaitoCommand

if TYPE_CHECKING:
    from aiogram.types import Message

router = Router(name="raito.system.stats")


@dataclass
class MemoryInformation:
    rss_mb: float
    vms_mb: float


@dataclass
class ProcessStats:
    uptime_sec: int
    memory: MemoryInformation
    cpu_percent: float


def get_process_stats() -> ProcessStats:
    proc = psutil.Process()

    with proc.oneshot():
        mem_info = proc.memory_info()
        cpu_percent = proc.cpu_percent(interval=0.1)
        create_time = proc.create_time()

    return ProcessStats(
        uptime_sec=int(time.time() - create_time),
        memory=MemoryInformation(
            rss_mb=mem_info.rss / 1024**2,
            vms_mb=mem_info.vms / 1024**2,
        ),
        cpu_percent=cpu_percent,
    )


def strf_seconds(seconds: int) -> str:
    days = seconds // 86400
    hours = (seconds % 86400) // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60

    return (
        (f"{days}d " if days else "")
        + (f"{hours}h " if hours else "")
        + (f"{minutes}m " if minutes else "")
        + (f"{seconds}s" if seconds else "")
    )


@router.message(RaitoCommand("stats"), DEVELOPER)
@description("Show process stats")
@hidden
async def stats(message: Message) -> None:
    stats = get_process_stats()

    text = "\n".join(
        [
            html.bold("Process stats"),
            "",
            f"• CPU: {stats.cpu_percent:.2f}%",
            f"• RAM: {stats.memory.rss_mb:.2f}mb",
            f"• Uptime: {strf_seconds(stats.uptime_sec)}",
        ]
    )
    await message.answer(text=text, parse_mode="HTML")



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, TypeVar

from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.fsm.context import FSMContext
from aiogram.types import Message
from typing_extensions import override

from raito.utils.helpers.filters import call_filters

from .registry import ConversationRegistry

if TYPE_CHECKING:
    from aiogram.types import TelegramObject

R = TypeVar("R")


__all__ = ("ConversationMiddleware",)


class ConversationMiddleware(BaseMiddleware):
    """Middleware for conversation handling."""

    def __init__(self, registry: ConversationRegistry) -> None:
        """Initialize ConversationMiddleware.

        :param registry: conversation registry
        """
        self.registry = registry

    @override
    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[R]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> R | None:
        """Process message with conversation support.

        :param handler: Next handler in the middleware chain
        :type handler: Callable
        :param event: Telegram event (Message)
        :type event: TelegramObject
        :param data: Additional data passed through the middleware chain
        :type data: dict[str, Any]
        :return: Handler result if not throttled, None if throttled
        """
        if not isinstance(event, Message):
            return await handler(event, data)

        context: FSMContext | None = data.get("state")
        if context is not None:
            state = await context.get_state()
            filters = self.registry.get_filters(context.key)

            if state is None or filters is None:
                return await handler(event, data)

            check = await call_filters(event, data, *filters)
            if not check:
                return await handler(event, data)

            self.registry.resolve(context.key, event)

        return await handler(event, data)



import asyncio
from collections.abc import Awaitable
from typing import TypeVar

from aiogram.exceptions import TelegramRetryAfter

T = TypeVar("T")


async def retry_method(
    func: Awaitable[T],
    max_attempts: int = 5,
    additional_delay: float = 0.1,
    *,
    _current_attempt: int = 1,
) -> T:
    """Retry a coroutine function if Telegram responds with RetryAfter.

    :param func: A coroutine function with no arguments.
    :param max_attempts: Maximum number of retry attempts.
    :param additional_delay: Extra seconds added to retry delay.
    :return: Result of the coroutine.
    :raises Exception: If all attempts fail or a non-RetryAfter exception is raised.
    """
    try:
        return await func
    except TelegramRetryAfter as exc:
        if _current_attempt >= max_attempts:
            raise

        await asyncio.sleep(exc.retry_after + additional_delay)
        return await retry_method(
            func,
            max_attempts,
            additional_delay,
            _current_attempt=_current_attempt + 1,
        )



from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING

from aiogram import Dispatcher, Router, html
from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.filters import Command, CommandObject
from aiogram.types import CallbackQuery, LinkPreviewOptions

from raito import Raito, rt
from raito.plugins.commands import description, hidden
from raito.plugins.pagination.enums import PaginationMode
from raito.plugins.pagination.paginators.list import ListPaginator
from raito.plugins.roles import DEVELOPER
from raito.utils.filters import RaitoCommand
from raito.utils.helpers.truncate import truncate

if TYPE_CHECKING:
    from aiogram.types import Message

router = Router(name="raito.help")


def _iter_raito_commands(
    dispatcher: Dispatcher,
) -> Generator[tuple[list[Command], HandlerObject], None, None]:
    for router in dispatcher.chain_tail:
        for handler in router.message.handlers:
            commands: list[Command] | None = handler.flags.get("commands")
            if handler.flags.get("raito__command") and commands:
                yield commands, handler


def _format_handler(command: Command, handler: HandlerObject) -> str:
    params: dict[str, type] = handler.flags.get("raito__params", {})
    params_str = " ".join(f"[{name}]" for name in params)

    signature = f"{command.prefix}{command.commands[0]} {params_str}"
    description: str = truncate(handler.flags.get("raito__description", ""), max_length=96)

    return html.code(signature) + "\n" + html.blockquote(html.italic(description))


def _get_formatted_commands(dispatcher: Dispatcher) -> list[str]:
    return [
        _format_handler(commands[0], handler)
        for commands, handler in _iter_raito_commands(dispatcher)
        if len(commands) > 0
    ]


@router.message(RaitoCommand("help"), DEVELOPER)
@description("Lists Raito commands")
@hidden
async def help_handler(message: Message, raito: Raito, command: CommandObject) -> None:
    if not message.bot or not message.from_user:
        return

    if command.args:
        for commands, handler in _iter_raito_commands(raito.dispatcher):
            if len(commands) <= 0:
                continue

            main_command = commands[0]
            if command.args in main_command.commands:
                await message.answer(_format_handler(main_command, handler), parse_mode="HTML")
                return

        await message.answer("⚠️ Command not found")
        return

    limit = 5
    raito_commands = _get_formatted_commands(raito.dispatcher)

    await raito.paginate(
        name="raito__commands",
        chat_id=message.chat.id,
        bot=message.bot,
        from_user=message.from_user,
        mode=PaginationMode.LIST,
        limit=limit,
        total_pages=ListPaginator.calc_total_pages(len(raito_commands), limit),
    )


@rt.on_pagination(router, "raito__commands", DEVELOPER)
async def on_pagination(
    _: CallbackQuery,
    raito: Raito,
    paginator: ListPaginator,
    offset: int,
    limit: int,
) -> None:
    commands = _get_formatted_commands(raito.dispatcher)
    footer = html.italic(html.link("Powered by Raito", "https://github.com/Aidenable/Raito"))

    await paginator.answer(
        items=[*commands[offset : offset + limit], footer],
        separator="\n\n",
        parse_mode="HTML",
        link_preview_options=LinkPreviewOptions(is_disabled=True),
    )



from __future__ import annotations

from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from typing import TYPE_CHECKING

from aiogram.dispatcher.event.handler import CallbackType
from aiogram.fsm.context import FSMContext
from aiogram.types import Message
from typing_extensions import Any

if TYPE_CHECKING:
    from raito import Raito

__all__ = ("Waiter", "wait_for")


@dataclass
class Waiter:
    """Container for conversation result.

    :param text: Raw text of the message
    :param number: Parsed integer if ``text`` is a digit, otherwise ``None``
    :param message: Original :class:`aiogram.types.Message` object
    :param retry: Callable coroutine for repeating the same wait
    """

    text: str
    number: int | None
    message: Message
    retry: Callable[[], Coroutine[Any, Any, Waiter]]


async def wait_for(
    raito: Raito,
    context: FSMContext,
    *filters: CallbackType,
) -> Waiter:
    """Wait for the next message from user that matches given filters.

    This function sets special state ``raito__conversation`` in FSM and
    suspends coroutine execution until user sends a message that passes
    all provided filters. Result is wrapped into :class:`Waiter`.

    :param raito: Current :class:`Raito` instance
    :param context: FSM context for current chat
    :param filters: Sequence of aiogram filters
    :return: Conversation result with text, parsed number and original message
    :raises RuntimeError: If handler object not found during filter execution
    :raises asyncio.CancelledError: If conversation was cancelled
    """
    await context.set_state(raito.registry.STATE)
    message = await raito.registry.listen(context.key, *filters)

    async def retry() -> Waiter:
        return await wait_for(raito, context, *filters)

    text = message.text or message.caption or ""
    return Waiter(
        text=text,
        number=int(text) if text.isdigit() else None,
        message=message,
        retry=retry,
    )



from typing import TypeVar

from aiogram.exceptions import TelegramBadRequest

T = TypeVar("T")


class SuppressNotModifiedError:
    """
    Context manager that suppresses the ``TelegramBadRequest`` exception
    with the message ``Bad Request: message is not modified``.

    This is useful when editing a Telegram message and the new content
    is identical to the existing one, which would otherwise raise
    an error from the Telegram API.

    Example:

        .. code-block:: python

            from raito.utils.errors import SuppressNotModifiedError

            with SuppressNotModifiedError():
                await message.edit_text("same text")

    :param ignore_message: The exact error message to match.
                           If this message is found in the raised
                           ``TelegramBadRequest``, the exception is suppressed.
                           Defaults to ``"Bad Request: message is not modified"``.
    :type ignore_message: str
    """

    def __init__(self, ignore_message: str = "Bad Request: message is not modified") -> None:
        """Initialize the context manager.

        :param ignore_message: Error message string to match against.
        """
        self.ignore_message = ignore_message

    def __enter__(self: T) -> T:
        """Enter the runtime context related to this object.

        :return: The context manager instance itself.
        """
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        _: object | None,
    ) -> bool:
        """Exit the runtime context and suppress the exception if it matches.

        :param exc_type: The exception type (if any).
        :param exc_val: The exception instance (if any).
        :param _: The traceback object (unused).
        :return: True if the exception should be suppressed, False otherwise.
        :rtype: bool
        """
        return (
            exc_type is TelegramBadRequest
            and exc_val is not None
            and self.ignore_message in str(exc_val)
        )



from .middleware import ConversationMiddleware
from .registry import ConversationRegistry
from .waiter import Waiter, wait_for

__all__ = (
    "ConversationMiddleware",
    "ConversationRegistry",
    "Waiter",
    "wait_for",
)



from __future__ import annotations

from asyncio import sleep
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, TypeVar

from aiogram.dispatcher.event.bases import REJECTED
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import Message
from cachetools import TTLCache
from typing_extensions import override

if TYPE_CHECKING:
    from aiogram.types import TelegramObject

R = TypeVar("R")


__all__ = ("AlbumMiddleware",)


class AlbumMiddleware(BaseMiddleware):
    """Middleware for album handling."""

    def __init__(
        self,
        delay: float | int = 0.6,
        max_size: int = 10_000,
    ) -> None:
        """Initialize AlbumMiddleware.

        :param flag_name: flag name to filter
        :type flag_name: str
        """
        self.delay = delay
        self._album_data = TTLCache[str, list[Message]](maxsize=max_size, ttl=delay * 5)

    @override
    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[R]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> R | None:
        """Process message with album support.

        :param handler: Next handler in the middleware chain
        :param event: Telegram event (Message or CallbackQuery)
        :param data: Additional data passed through the middleware chain
        :return: Handler result
        """
        if not isinstance(event, Message):
            return await handler(event, data)

        if not event.media_group_id:
            return await handler(event, data)

        if album_data := self._album_data.get(event.media_group_id):
            album_data.append(event)
            return REJECTED

        self._album_data[event.media_group_id] = [event]
        await sleep(self.delay)

        # after sending all media files:
        data["album"] = self._album_data.pop(event.media_group_id)

        return await handler(event, data)



from __future__ import annotations

import ast
import sys
import traceback
from collections.abc import Awaitable, Callable, Generator
from contextlib import contextmanager, redirect_stdout
from dataclasses import dataclass
from io import StringIO
from typing import Any


@dataclass
class EvaluationData:
    """Stores the result of evaluated python code.

    :param stdout: Captured standard output (e.g. from ``print()``)
    :param result: The final returned result.
    :param error: Traceback string if an exception occurred.
    """

    stdout: str | None = None
    result: str | None = None
    error: str | None = None


class CodeEvaluator:
    """Async code evaluator with stdout capture and error handling."""

    @contextmanager
    def _capture_output(self) -> Generator[StringIO, Any, None]:
        """Context manager that captures ``stdout`` into a buffer.

        :yield: A ``StringIO`` buffer with captured output.
        """
        old_stdout = sys.stdout
        buf = StringIO()

        try:
            with redirect_stdout(buf):
                yield buf
        finally:
            sys.stdout = old_stdout

    def _wrap_code(self, code: str) -> ast.Module:
        """Wraps code into an async function, returning the last expression.

        :param code: python code to wrap.
        :return: A module with a single async def.
        """
        module = ast.parse(code, mode="exec")
        body = module.body or [ast.Pass()]

        if isinstance(body[-1], ast.Expr):
            body[-1] = ast.Return(value=body[-1].value)

        func_def = ast.AsyncFunctionDef(
            name="__eval_func",
            args=ast.arguments(
                posonlyargs=[],
                args=[],
                defaults=[],
                kwonlyargs=[],
                kw_defaults=[],
            ),
            body=body,
            decorator_list=[],
        )

        wrapped_module = ast.Module(body=[func_def], type_ignores=[])
        ast.fix_missing_locations(wrapped_module)
        return wrapped_module

    async def evaluate(self, code: str, context: dict[str, Any]) -> EvaluationData:
        """Evaluates the given async python code with the provided context.

        :param code: Python code to execute.
        :param context: Variables available during execution.
        :return: Result of the evaluation as ``EvaluationData``
        """
        try:
            wrapped_module = self._wrap_code(code)
            compiled_code = compile(wrapped_module, "<raito_eval>", "exec")

            exec_locals: dict[str, Any] = {}
            with self._capture_output() as output:
                exec(compiled_code, context, exec_locals)
                eval_func: Callable[[], Awaitable[Any]] = exec_locals["__eval_func"]
                result = await eval_func()

            return EvaluationData(
                stdout=output.getvalue(),
                result=str(result) if result is not None else None,
            )
        except Exception:  # noqa: BLE001
            return EvaluationData(error=traceback.format_exc())



from __future__ import annotations

from html import escape
from typing import TYPE_CHECKING, Any

from aiogram import F, Router, html
from aiogram.filters import CommandObject
from aiogram.fsm.state import State, StatesGroup

from raito.plugins.commands import description, hidden
from raito.plugins.roles import DEVELOPER
from raito.utils.filters import RaitoCommand
from raito.utils.helpers.code_evaluator import CodeEvaluator

if TYPE_CHECKING:
    from aiogram.fsm.context import FSMContext
    from aiogram.types import Message

router = Router(name="raito.system.eval")
code_evaluator = CodeEvaluator()


class EvalGroup(StatesGroup):
    expression = State()


async def _execute_code(message: Message, code: str, data: dict[str, Any]) -> None:
    data = {"_" + k: v for k, v in data.items()}
    data["_msg"] = message
    data["_user"] = message.from_user

    evaluation_data = await code_evaluator.evaluate(code, data)
    pre_blocks = []

    if evaluation_data.stdout:
        pre_blocks.append(evaluation_data.stdout[:1000])

    if evaluation_data.error:
        pre_blocks.append(evaluation_data.error[:3000])
    elif evaluation_data.result is not None:
        pre_blocks.append(evaluation_data.result[:3000])
    else:
        pre_blocks.append("no output")

    text = "\n\n".join([html.pre(escape(i)) for i in pre_blocks])
    await message.answer(text=text, parse_mode="HTML")


@router.message(RaitoCommand("eval", "py", "py3", "python", "exec"), DEVELOPER)
@description("Execute Python script")
@hidden
async def eval_handler(
    message: Message,
    state: FSMContext,
    command: CommandObject,
    **data: Any,  # noqa: ANN401
) -> None:
    if not command.args:
        await message.answer(text="📦 Enter Python expression:")
        await state.set_state(EvalGroup.expression)
        return

    data["message"] = message
    data["state"] = state
    data["command"] = command
    await _execute_code(message, command.args, data)


@router.message(EvalGroup.expression, F.text, DEVELOPER)
async def eval_process(
    message: Message,
    state: FSMContext,
    **data: Any,  # noqa: ANN401
) -> None:
    await state.clear()

    if not message.text:
        await message.answer(text="⚠️ Expression cannot be empty")
        return

    data["message"] = message
    data["state"] = state
    await _execute_code(message, message.text, data)
